@@ -114,7 +114,7 @@ ngx_copy(u_char *dst, u_char *src, size_t len)
 
 
 /* msvc and icc7 compile memcmp() to the inline loop */
-#define ngx_memcmp                memcmp
+#define ngx_memcmp(s1, s2, n)  memcmp((const char *) s1, (const char *) s2, n)
 
 
 u_char *ngx_cpystrn(u_char *dst, u_char *src, size_t n);
@@ -141,7 +141,7 @@ off_t ngx_atoof(u_char *line, size_t n);
 time_t ngx_atotm(u_char *line, size_t n);
 ngx_int_t ngx_hextoi(u_char *line, size_t n);
 
-void ngx_md5_text(u_char *text, u_char *md5);
+u_char *ngx_hex_dump(u_char *dst, u_char *src, size_t len);
 
 
 #define ngx_base64_encoded_length(len)  (((len + 2) / 3) * 4)
@@ -173,7 +173,7 @@ uintptr_t ngx_escape_html(u_char *dst, u_char *src, size_t size);
 
 
 void ngx_sort(void *base, size_t n, size_t size,
-    int (*cmp)(const void *, const void *));
+    ngx_int_t (*cmp)(const void *, const void *));
 #define ngx_qsort             qsort
 
 
