@@ -230,6 +230,7 @@ typedef struct {
     ngx_table_elt_t                  *content_length;
     ngx_table_elt_t                  *content_encoding;
     ngx_table_elt_t                  *location;
+    ngx_table_elt_t                  *refresh;
     ngx_table_elt_t                  *last_modified;
     ngx_table_elt_t                  *content_range;
     ngx_table_elt_t                  *accept_ranges;
@@ -372,7 +373,6 @@ struct ngx_http_request_s {
     uint32_t                          in_addr;
     ngx_uint_t                        port;
     ngx_str_t                        *port_text;    /* ":80" */
-    ngx_str_t                         server_name;
     ngx_http_virtual_names_t         *virtual_names;
 
     ngx_int_t                         phase_handler;
@@ -427,7 +427,7 @@ struct ngx_http_request_s {
     unsigned                          fast_subrequest:1;
     unsigned                          subrequest_in_memory:1;
 
-    unsigned                          header_timeout_set:1;
+    unsigned                          gzip:2;
 
     unsigned                          proxy:1;
     unsigned                          bypass_cache:1;
